/* heur_tope.c: Simple zonotope-based Q01P heuristic
 * idea Tom Liebling
 * written Sat Sep 29 16:54:53 2001 par Jean-Albert Ferrez
 * based on rs_tope.c by Jean-Albert Ferrez and Komei Fukuda
 * based on code from Kim Allemand (rsz.c)
 * and Komei Fukuda (rs_knapsack.c)
 * with many parts borrowed from rs_example.c by Ambros Marzetta
 * Last modified on July 13, 2002 by Komei Fukuda to be compatible with cddlib-092a.
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "z_base.h"
#include "z_rs.h"
#include "z_io.h"
#include "z_time.h"

#ifndef __SETOPER_H
#include "setoper.h"
#endif

#ifndef __CDD_H
#include "cdd.h"
#endif

/* typedef char boolean;*/  /* used to be provided by ZRAM */

typedef struct rst_Node {
  int m; /* the number of halfspaces (inequalities) */
  int d; /* the dimension of the central arrangement */
  short x[1]; /* the sign vector of size m to represent a tope */
} rst_Node, *rst_NodeP;

typedef struct rst_Global {
  int m; /* the number of halfspaces (inequalities) */
  int d; /* the dimension of the central arrangement */
  double bestval; /* storing the current best value */
  signed long mat[1]; 
    /* a (d+1) x m matrix to represent the m x d input matrix AND the 
       location vector x[] of the initial tope.   
       The columns of the input matrix is stored consecutively and
       the last m elements represent x[] vector.  */
} rst_Global, *rst_GlobalP;

void rst_Print(rst_GlobalP, rst_NodeP);
boolean rst_IsRoot(rst_GlobalP, rst_NodeP);

/* Static variables (initialized on every processor by initproc() */

/* Mandatory application-specific procedures ------------------ */

rst_NodeP rst_NewNode(long m, long d)
{
  rst_NodeP new;

  new = (rst_NodeP) malloc(sizeof(rst_Node)+m*sizeof(short));
  new->m=m;
  new->d=d;
  assert(new);
  return new;
} /* rst_NewNode */

void rst_FreeNode(rst_NodeP node)
{
  assert(node);
  free(node);
} /* rst_FreeNode */

boolean rst_Equal(rst_GlobalP global, rst_NodeP a, rst_NodeP b)
{
  long j=0;
  boolean answer=1;

  assert(a);
  assert(b);
  while (answer && j < global->m){
    if (a->x[j]!=b->x[j]) answer=0;  /* two nodes are distinct */
    j++;
  }
  return answer;
} /* rst_Equal */

long rst_Delta(rst_GlobalP global, rst_NodeP node)
{
  return (global->m);
} /* rst_Delta */


long AdjOracle(rst_GlobalP global, rst_NodeP node, long k)
{
  long m, d;
  dd_boolean res;
  dd_rowrange i;
  dd_colrange j;
  dd_MatrixPtr M;
  dd_Arow cvec=NULL;  /* certificate of nonredundancy */
  dd_LPPtr lp;
  dd_ErrorType err;
  dd_LPSolutionPtr lps;
  dd_boolean localdebug=dd_FALSE;


  assert(node);
  assert(k>0);

  /*  Defer  */
  m = global->m;
  d = global->d;
  M = dd_CreateMatrix(m,d+1);
  dd_InitializeArow(d+1, &cvec);

  /*  Fill the constraints for the given node */
  for (i=0; i<m; i++)
    {
      dd_set_si(M->matrix[i][0], 0);  /* central arrangement */
      if (node->x[i]) 
        for (j=1; j<=d; j++) dd_set_si(M->matrix[i][j],  global->mat[i+(j-1)*m]);
      else
        for (j=1; j<=d; j++) dd_set_si(M->matrix[i][j], -global->mat[i+(j-1)*m]);
    }

  /* Set the first constraint to be non homogeneous ( >=  1),
  assuming that the initial tope is on the positive side of the first.  */
  dd_set_si(M->matrix[0][0], -1);
  M->representation=dd_Inequality;
  
  /*  Checking if the k+1 st constraint is redundant.  */
  res=dd_Redundant(M, k+1, cvec, &err);
  if (localdebug) {
    fprintf(stderr, "\nRedundancy checking for k=%ld\n",k);
    rst_Print(global,node);
    /* dd_WriteMatrix(stderr,M); */
    if (res) fprintf(stderr, "\nit is redundant.\n");
    else {
      fprintf(stderr, "\nit is nonredundant.\n");
      /* dd_WriteArow(stderr, cvec, d+1);  */
    } 
  }
  
  if (err!=dd_NoError) {
    fprintf(stderr, "Warning: a redundancy checking failed\n");
    dd_WriteErrorMessages(stderr, err);
  }
  
  /* For debugging, write the matrix */
  /* dd_WriteMatrix(stderr, M);  */
  
  dd_FreeMatrix(M);
  dd_FreeArow(d+1, cvec);
  
  if (res) return 0;   /* the k+1 st constraint is redundant */
  return 1;    /* the k+1 st constraint is nonredundant */
  
} /* AdjOracle */


long ShootingLocalSearch(rst_GlobalP global, rst_NodeP node)
{
  /* This is a new local search using the canonical interior point of
  each tope.  The next tope is the one the line connecting the canonical
  tope and that of the root tope.
  */
  long c;
  long m, d;
  dd_boolean res;
  dd_rowrange i;
  dd_colrange j;
  dd_Arow rootv=NULL, nodev=NULL, direction=NULL;
  dd_MatrixPtr M;
  dd_LPPtr lp0, lp;
  dd_LPSolutionPtr lps;
  dd_ErrorType err;
  dd_LPSolverType solver=dd_DualSimplex;
  dd_boolean localdebug=dd_FALSE;

  assert(node);

  if (rst_IsRoot(global, node)) return 0;

  /*  Set up the matrix for the root.  */
  m = global->m;
  d = global->d;
  M = dd_CreateMatrix(m,d+1);
  dd_InitializeArow(d+1, &rootv);
  dd_InitializeArow(d+1, &nodev);
  dd_InitializeArow(d+1, &direction);

  /*  Fill the constraints for the root node */
  for (i=0; i<m; i++)
    {
      dd_set_si(M->matrix[i][0], 0);  /* central arrangement */
      for (j=1; j<=d; j++) dd_set_si(M->matrix[i][j], global->mat[i+(j-1)*m]);
    }

  /* Set the first constraint to be non homogeneous ( >=  1),
  assuming that the initial tope is on the positive side of the first.  */
  dd_set_si(M->matrix[0][0], -1);
  M->representation=dd_Inequality;
  if (localdebug){ 
    fprintf(stderr, "The input matrix is set to:\n");
    dd_WriteMatrix(stderr, M);
  }

  for (j=0; j<=d; j++) dd_set_si(M->rowvec[j],0);  /* this is not important for our purpose. */
  M->objective=dd_LPmax;      /* this is not important for our purpose. */
  lp0=dd_Matrix2LP(M, &err);  assert(lp0);
  lp=dd_MakeLPforInteriorFinding(lp0);
  dd_LPSolve(lp,solver,&err);
  lps=dd_CopyLPSolution(lp);
  if (dd_Nonpositive(lps->optvalue)) {
    fprintf(stderr, "ShootingLocalSearch: Error.  No interior point was found for the root!!!\n");
    return 0;
  }
  for (j=0; j<=d; j++) dd_set(rootv[j],lps->sol[j]);  /* copy an interior point of the root tope */
  if (localdebug){
    fprintf(stderr,"An interior point of the root node:\n");
    dd_WriteArow(stderr, rootv, d+1);  fprintf(stderr,"\n");
  }
  dd_FreeLPData(lp0); dd_FreeLPData(lp);
  dd_FreeLPSolution(lps);

  /*  Set the constraints for the current node */
  for (i=0; i<m; i++)
    if (node->x[i]<=0) /* reverse the inequality */
      for (j=0; j<=d; j++) dd_neg(M->matrix[i][j], M->matrix[i][j]);
  lp0=dd_Matrix2LP(M, &err);  assert(lp0);
  lp=dd_MakeLPforInteriorFinding(lp0);
  dd_LPSolve(lp,solver,&err);
  lps=dd_CopyLPSolution(lp);
  if (dd_Nonpositive(lps->optvalue)) {
    fprintf(stderr, "ShootingLocalSearch: Error.  No interior point was found for the node!!!\n");
    printf("ShootingLocalSearch: Error.  No interior point was found for the node!!!\n");
    printf("About to print node and corresponding LP.");
    rst_Print(global,node);
    dd_WriteLP(stderr, lp); 
    printf("Printed node and corresponding LP.");
    fflush(stdout);
    return 0;
  }
  for (j=0; j<=d; j++) dd_set(nodev[j],lps->sol[j]);  /* copy an interior point of the current tope */
  if (localdebug){
    fprintf(stderr,"An interior point of the current node:\n");
    dd_WriteArow(stderr, nodev, d+1);  fprintf(stderr,"\n");
  }

  for (j=0; j<=d; j++) dd_sub(direction[j],rootv[j],nodev[j]);  /* the direction of shooting. */
  if (localdebug){
    fprintf(stderr,"Rayshooting is performed for the local search from:\n");
    dd_WriteArow(stderr, nodev, d+1);  fprintf(stderr,"to:\n");
    dd_WriteArow(stderr, rootv, d+1);  fprintf(stderr,"with direction:\n");
    dd_WriteArow(stderr, direction, d+1);  fprintf(stderr,"\n with matrix\n");
    dd_WriteMatrix(stderr, M);
  }

  c=dd_RayShooting(M, nodev, direction);  /* Ray shooting from nodev towards rootv.

  printf("Rayshooting hits the row  %ld\n", c);

  /* free the allocated space */
  dd_FreeMatrix(M);
  dd_FreeArow(d+1, rootv);
  dd_FreeArow(d+1, nodev);
  dd_FreeArow(d+1, direction);
  dd_FreeLPData(lp0); dd_FreeLPData(lp);
  dd_FreeLPSolution(lps);

  return c-1; /* Normal termination */
} /* ShootingLocalSearch */


long MinIndexLocalSearch(rst_GlobalP global, rst_NodeP node)
{
  long c;

  /*  Explore all neighbourd (except [0])  */
  for (c=1; c<global->m; c++)
    {   
      /*  Execute oracle  */
      if (node->x[c]<=0 && AdjOracle(global, node, c)) 
        {
          return c;
        }
    }
  return 0; /* as 0 can never be candidate */
} /* MinIndexLocalSearch */


rst_NodeP rst_AdjOracle(rst_GlobalP global, rst_NodeP node, long k)
{
  int LPtestRes;
  long i;

  rst_NodeP valid;
  rst_NodeP result;

  assert(node);

  /* No need to check on 0th index */
  if (k == 0 || k>=global->m) return NULL;

  
  /* No need to check if node is on the negative side  */
  if (node->x[k] <= 0) return NULL;


  LPtestRes = AdjOracle(global, node, k);

  if (LPtestRes != 1)
    {
/*       printf("rst_AdjOracle: empty cell, dropping\n"); */
      return NULL;
    }

  /*  New neighbour candidate has been found  */
  result = rst_NewNode(global->m,global->d);
  for (i=0; i<global->m; i++) result->x[i]=node->x[i];  /* copy the node first. */
  if (result->x[k]) result->x[k]=0; else result->x[k]=1;  /* move to the other side of the (k+1)st hyperplane. */
  assert(result);

  /* printf("rst_AdjOracle: %ld th neighbor found\n", k); */
  return result;

} /* rst_AdjOracle */


boolean rst_IsRoot(rst_GlobalP global, rst_NodeP node)
{
  long i=0;
  boolean answer=1;

  assert(node);
  assert(global);

  while (answer && i < global->m){
    if (node->x[i]!=1) answer=0;  /* the node is not the root.  ??? 0 or 1 */
    i++;
  }
  return answer;
} /* rst_IsRoot */


rst_NodeP rst_F(rst_GlobalP global, rst_NodeP node)
{
  long LS,i;
  rst_NodeP result;
  dd_boolean localdebug=dd_FALSE;

  assert(node);

  if (localdebug) {
    fprintf(stderr, "LocalSearch function is applied to the node\n", LS);
    rst_Print(global,node);
  }
#if defined MINLS
  LS = MinIndexLocalSearch(global, node);
#else
  LS = ShootingLocalSearch(global, node);
#endif
  if (localdebug)
    fprintf(stderr, "LocalSearch function returns %ld\n", LS);
  if (LS == 0) return NULL;
  else
    {
      result = rst_NewNode(global->m, global->d);
      for (i=0; i<global->m; i++) result->x[i]=node->x[i];  /* copy the node first. */
      if (result->x[LS]) result->x[LS]=0; else result->x[LS]=1;  /* move to the other side. */
    }
  return result;

} /* rst_F */

/* -------------------------------------------------------
    UP TO HERE for KOMEI.
   -------------------------------------------------------
*/


double rst_Evaluate(rst_GlobalP global, rst_NodeP node)
{
  int im, id;
  double res, t;       /* HUGE PROBLEM wrt overflows */
  long *origcell;

  origcell = &(global->mat[global->m*global->d]);  /* alias pointer */

  /*  calcul "brutal", peut jtre optimisi   */
  res = 0.0;
  for (id=0; id<global->d; id++)
    {
      t = 0.0;
      for (im=0; im<global->m; im++)
        {
          if (node->x[im] == 0 && origcell[im] == 1)
            t -= global->mat[global->m*id+im];
          if (node->x[im] == 1 && origcell[im] == 0)
            t += global->mat[global->m*id+im];
        }
      t *= t;
      res += t;
    }
  if (global->bestval < res) global->bestval=res; /* update the best value */
  return res;
}

void rst_Print(rst_GlobalP global, rst_NodeP node)
{
  int im;
  long *origcell;
  double val;

  origcell = &(global->mat[global->m*global->d]);  /* alias pointer */

  /*  Current cell  */
  val=rst_Evaluate(global, node);
  printf("%10.4e: ", val);
  if (val >= global->bestval){
    for (im=0; im<global->m; im++)
      if (node->x[im] == 1)
        printf(origcell[im] == 0 ? "1" : "0");
      else
        printf(origcell[im] == 0 ? "0" : "1");
  }
  printf("\n");

  /*  "Symetric" cell  */
/*   for (im=0; im<global->m; im++) */
/*     node->x[im] = 1 - node->x[im]; */

/*   val=rst_Evaluate(global, node); */
/*   printf("%10.4e: ", val);                        */
/*   if (val >= global->bestval){ */
/*     for (im=0; im<global->m; im++) */
/*       if (node->x[im] == 1) */
/*         printf(origcell[im] == 0 ? "1" : "0"); */
/*       else */
/*         printf(origcell[im] == 0 ? "0" : "1"); */
/*   } */
/*   printf("\n"); */

  /*  Revert to current cell  */
/*   for (im=0; im<global->m; im++) */
/*     node->x[im] = 1 - node->x[im]; */

} /* rst_Print */




int main(int argc, char **argv, char **envp)
{
  int m, d, im, id, j, step;
  long tmp;

  rst_GlobalP global;
  rst_NodeP root;

  double stime,ttime;
  double curval, newval;

  int found;
  long *testpoint, *origcell;

  char fname[250];
  FILE *iofile;

  if (argc != 2)
    {
      fprintf(stderr, "Usage: %s problemname\n", argv[0]);
      fprintf(stderr, "  reads problemname.input\n", argv[0]);
      exit(2);
    }

  snprintf(fname, 250, "%s.input", argv[1]);
  iofile = fopen(fname, "r");

  dd_set_global_constants();

  /*------------------------------------------------------------------*/
  /*  Read objective function                                         */
  /*------------------------------------------------------------------*/
  fscanf(iofile, "%d %d", &m, &d);

  global = (rst_GlobalP) malloc(sizeof(rst_Global) + m*(d+1)*sizeof(long));
  global->m = m;
  global->d = d;

  for (im=0; im<m; im++)
    for (id=0; id<d; id++)
      { /*  can read only integers  */
        if (fscanf(iofile, "%ld", &tmp)!=1)
          {
            fprintf(stderr, 
                    "rs_tope: Error while reading objective function\n");
            exit(1);
          }
        global->mat[m*id+im] = tmp;
      }

  fclose(iofile);

  iofile = stdout;
/*   snprintf(fname, 250, "%s.output", argv[1]); */
/*   iofile = fopen(fname, "w"); */
/*   snprintf(fname, 250, "%s.output-PE", argv[1]); */

  fprintf(iofile, "Done reading. Objective function: m=%d, d=%d\n", m, d);

  /*  Find first cell. Very naive but sufficient  */
  testpoint = malloc(d*sizeof(long));
  origcell = &(global->mat[m*d]);   /* alias pointer */


  for (j=1; j<11; j++) {

  found = 0;
  while (!found)
    {
      for (id=0; id<d; id++) 
        testpoint[id] = rand()%32000-16000;
      for (im=0; im<m; im++)
        {
          tmp = 0;
          for (id=0; id<d; id++)
            tmp += testpoint[id] * global->mat[m*id+im];
          if (tmp == 0)
            break;
          if (im == 0 && tmp < 0)  /* force [0]=1 */
            break;
          origcell[im] = (tmp > 0 ? 1 : 0);
        }
      if (im == m)
        {
          found = 1;  /* Got our root cell */
          /*  Invert matrix members such that the signature
              of the root cell is all 1 */
          for (im=0; im<m; im++)
            if (origcell[im] == 0) 
              for (id=0; id<d; id++)
                global->mat[m*id+im] *= -1;
         }
    }

/*   fprintf(iofile, "The initial point :"); */
/*   for (id=0; id<d; id++) */
/*     fprintf(iofile, " %ld ", testpoint[id]); */
/*   fprintf(iofile, "\n"); */
  /* free(testpoint); */

  root = (rst_NodeP) rst_NewNode(m, d);
  root->m = m;
  root->d = d;
  for (im=0; im<m; im++)
    root->x[im] = 1;

  global->bestval=0; /* the current best value */
  curval = rst_Evaluate(global, root);
/*   fprintf(iofile, "Starting with %10.4e\n", curval); */
/*   rst_Print(global, root); */
  step = 0;

  found = 0;    /* found means here I got stuck in a local opt... */
  while (!found)
    {
      for (im=0; im<m; im++)
        {
          if (AdjOracle(global, root, im) == 1)
            {
              root->x[im] = 1 - root->x[im];
              newval = rst_Evaluate(global, root);
              if (curval < newval)
                {
                  curval = newval;
/*                   fprintf(iofile, "Found better : %10.4e\n", newval); */
/*                   rst_Print(global, root); */
                  step++;
                  break;
                } else {
                  root->x[im] = 1 - root->x[im];
                }
            }
/*           fprintf(stderr, "."); */
        }
      if (im == m)
        {
          found = 1; /* no more better adj cell, stopping */
          printf("%2d %2d ", j, step);
          rst_Print(global, root);
        }
    }

  /*------------------------------------------------------------------*/
  /*  Desallocation                                                   */
  /*------------------------------------------------------------------*/

  /* Should free some stuff... */

  /* reinvert values of the matrix for next step... */
          for (im=0; im<m; im++)
            if (origcell[im] == 0) 
              for (id=0; id<d; id++)
                global->mat[m*id+im] *= -1;
  }

  fprintf(iofile, "End of Zonotopes Q01P heuristic\n");
/*   fclose(iofile); */
  return 0;

} /* mmain */

